# SPDX-FileCopyrightText: Copyright (c) 2020 Bryan Siepert for Adafruit Industries
#
# SPDX-License-Identifier: MIT

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
`analog_in`
==============================
AnalogIn for ADC readings.

* Author(s): Bryan Siepert, adpted from ADS1x15 by Carter Nelson
"""


class AnalogIn:
    """AnalogIn Mock Implementation for ADC Reads."""

    def __init__(self, pcf, pin):
        """AnalogIn

        :param ads: The PCF8591 object.
        :param ~digitalio.DigitalInOut pin: Required ADC channel pin.

        """
        self._pcf = pcf
        self._channel_number = pin

    @property
    def voltage(self):
        """Returns the value of an ADC channel in volts as compared to the reference voltage."""

        if not self._pcf:
            raise RuntimeError(
                "Underlying ADC does not exist, likely due to callint `deinit`"
            )
        raw_reading = self._pcf.read(self._channel_number)
        return ((raw_reading << 8) / 65535) * self._pcf.reference_voltage

    @property
    def value(self):
        """Returns the value of an ADC channel.
        The value is scaled to a 16-bit integer from the native 8-bit value."""

        if not self._pcf:
            raise RuntimeError(
                "Underlying ADC does not exist, likely due to callint `deinit`"
            )

        return self._pcf.read(self._channel_number) << 8

    @property
    def reference_voltage(self):
        """The maximum voltage measurable (also known as the reference voltage) as a float in
        Volts. Assumed to be 3.3V but can be overridden using the `PCF8591` constructor"""
        if not self._pcf:
            raise RuntimeError(
                "Underlying ADC does not exist, likely due to callint `deinit`"
            )
        return self._pcf.reference_voltage

    def deinit(self):
        """Release the reference to the PCF8591. Create a new AnalogIn to use it again."""
        self._pcf = None



# SPDX-FileCopyrightText: 2017 Scott Shawcroft, written for Adafruit Industries
# SPDX-FileCopyrightText: Copyright (c) 2020 Bryan Siepert for Adafruit Industries
#
# SPDX-License-Identifier: MIT
"""
`pcf8591`
================================================================================

ADC+DAC Combo

* Author(s): Bryan Siepert

Implementation Notes
--------------------

**Hardware:**

* `Adafruit PCF8591 Breakout <https://www.adafruit.com/products/45XX>`_

**Software and Dependencies:**

* Adafruit CircuitPython firmware for the supported boards:
  https://github.com/adafruit/circuitpython/releases

 * Adafruit's Bus Device library: https://github.com/adafruit/Adafruit_CircuitPython_BusDevice
 * Adafruit's Register library: https://github.com/adafruit/Adafruit_CircuitPython_Register
"""

__version__ = "0.0.0-auto.0"
__repo__ = "https://github.com/adafruit/Adafruit_CircuitPython_PCF8591.git"
# from time import sleep
from struct import unpack_from
from micropython import const
from adafruit_bus_device import i2c_device

_PCF8591_DEFAULT_ADDR = const(0x48)  # PCF8591 Default Address
_PCF8591_ENABLE_DAC = const(0x40)  # control bit for having the DAC active

# Pin constants
A0 = const(0)
A1 = const(1)
A2 = const(2)
A3 = const(3)

OUT = const(0)


class PCF8591:
    """Driver for the PCF8591 DAC & ADC Combo breakout.

    :param ~busio.I2C i2c_bus: The I2C bus the PCF8591 is connected to.
    :param address: The I2C device address for the sensor. Default is ``0x28``.

    """

    def __init__(self, i2c_bus, address=_PCF8591_DEFAULT_ADDR, reference_voltage=3.3):
        self.i2c_device = i2c_device.I2CDevice(i2c_bus, address)
        self._dacval = 0
        self._dac_enabled = False
        # this is the range supported by the PCF8591
        if 2.5 <= reference_voltage <= 6.0:
            self._reference_voltage = reference_voltage
        else:
            raise ValueError("reference_voltage must be from 2.5 - 6.0")
        self._buffer = bytearray(2)
        # possibly measure each channel here to prep readings for
        # user calls to `read`

    @property
    def reference_voltage(self):
        """The voltage level that ADC signals are compared to.
        An ADC value of 65535 will equal `reference_voltage`"""
        return self._reference_voltage

    def _half_read(self, channel):
        if self._dac_enabled:
            self._buffer[0] = _PCF8591_ENABLE_DAC
            self._buffer[1] = self._dacval
        else:
            self._buffer[0] = 0
            self._buffer[1] = 0

        self._buffer[0] |= channel & 0x3

        with self.i2c_device as i2c:
            i2c.write_then_readinto(self._buffer, self._buffer)

    def read(self, channel):
        """Read an analog value from one of the four ADC inputs

        param: :channel The single-ended ADC channel to read from, 0 thru 3
        """
        if channel < 0 or channel > 3:
            raise ValueError("channel must be from 0-3")
        # reads are started on the ACK of the WRITE to the 'register' and
        # not returned until the read after the _next_ WRITE so we have to
        # do it twice to get the actual value
        self._half_read(channel)
        self._half_read(channel)

        return unpack_from(">B", self._buffer[1:])[0]

    @property
    def dac_enabled(self):
        """ Enables the DAC when True, or sets it to tri-state / high-Z when False"""
        return self._dac_enabled

    @dac_enabled.setter
    def dac_enabled(self, enable_dac):

        self._dac_enabled = enable_dac
        self.write(self._dacval)

    def write(self, value):
        """Writes a uint8_t value to the DAC output

        param: :output The value to write: 0 is GND and 65535 is VCC

        """

        self._buffer = bytearray(2)
        if self._dac_enabled:
            self._buffer[0] = _PCF8591_ENABLE_DAC
            self._buffer[1] = value
        self._dacval = value
        with self.i2c_device as i2c:
            i2c.write_then_readinto(self._buffer, self._buffer)



# SPDX-FileCopyrightText: Copyright (c) 2020 Bryan Siepert for Adafruit Industries
# SPDX-License-Identifier: MIT
import time
import board

import adafruit_pcf8591.pcf8591 as PCF
from adafruit_pcf8591.analog_in import AnalogIn
from adafruit_pcf8591.analog_out import AnalogOut

############# AnalogOut & AnalogIn Example ##########################
#
# This example shows how to use the included AnalogIn and AnalogOut
# classes to set the internal DAC to output a voltage and then measure
# it with the first ADC channel.
#
# Wiring:
# Connect the DAC output to the first ADC channel, in addition to the
# normal power and I2C connections
#
#####################################################################
i2c = board.I2C()
pcf = PCF.PCF8591(i2c)

pcf_in_0 = AnalogIn(pcf, PCF.A0)
pcf_out = AnalogOut(pcf, PCF.OUT)

while True:

    print("Setting out to ", 65535)
    pcf_out.value = 65535
    raw_value = pcf_in_0.value
    scaled_value = (raw_value / 65535) * pcf_in_0.reference_voltage

    print("Pin 0: %0.2fV" % (scaled_value))
    print("")
    time.sleep(1)

    print("Setting out to ", 32767)
    pcf_out.value = 32767
    raw_value = pcf_in_0.value
    scaled_value = (raw_value / 65535) * pcf_in_0.reference_voltage

    print("Pin 0: %0.2fV" % (scaled_value))
    print("")
    time.sleep(1)

    print("Setting out to ", 0)
    pcf_out.value = 0
    raw_value = pcf_in_0.value
    scaled_value = (raw_value / 65535) * pcf_in_0.reference_voltage

    print("Pin 0: %0.2fV" % (scaled_value))
    print("")
    time.sleep(1)



# SPDX-FileCopyrightText: 2017 Scott Shawcroft, written for Adafruit Industries
# SPDX-FileCopyrightText: Copyright (c) 2020 Bryan Siepert for Adafruit Industries
#
# SPDX-License-Identifier: MIT
import time
import board
from adafruit_pcf8591.pcf8591 import PCF8591

################ read/DAC Example #####################
#
# This example shows how to use a PCF8591 instance to set the voltage output by the included DAC.
#
# Wiring:
# Connect the DAC output to the first ADC channel, in addition to the
# normal power and I2C connections
#
########################################
i2c = board.I2C()

pcf = PCF8591(i2c)
print("enabling DAC")
pcf.dac_enabled = True

while True:
    print("Setting DAC to", 255)
    pcf.write(255)

    read_value = pcf.read(0)
    scaled_value = (read_value / 255) * pcf.reference_voltage
    print("Channel 0: %0.2fV" % (scaled_value))
    print("")
    time.sleep(0.2)

    print("Setting DAC to", 127)
    pcf.write(127)

    read_value = pcf.read(0)
    scaled_value = (read_value / 255) * pcf.reference_voltage
    print("Channel 0: %0.2fV" % (scaled_value))
    print("")
    time.sleep(0.2)

    print("Setting DAC to", 0)
    pcf.write(0)

    read_value = pcf.read(0)
    scaled_value = (read_value / 255) * pcf.reference_voltage
    print("Channel 0: %0.2fV" % (scaled_value))
    print("")
    time.sleep(0.2)



# SPDX-FileCopyrightText: Copyright (c) 2020 Bryan Siepert for Adafruit Industries
#
# SPDX-License-Identifier: MIT
import time
import board
from adafruit_pcf8591.pcf8591 import PCF8591

################ Read/ADC Example #####################
#
# This example shows how to use a PCF8591 instance to read one of the ADC channels.
#
# Wiring:
# Connect a voltage source to the first ADC channel, in addition to the
# normal power and I2C connections. The voltage level should be between 0V/GND and VCC
#
########################################
i2c = board.I2C()
pcf = PCF8591(i2c)

channel_a = 0
channel_b = 1

while True:

    read_value = pcf.read(channel_a)
    scaled_value = (read_value / 255) * pcf.reference_voltage

    print("Channel: %d %0.2fV" % (channel_a, scaled_value))
    print("")
    time.sleep(0.1)

    read_value = pcf.read(channel_b)
    scaled_value = (read_value / 255) * pcf.reference_voltage

    print("Channel: %d %0.2fV" % (channel_b, scaled_value))
    print("")
    print("*" * 20)
    time.sleep(1)



# SPDX-FileCopyrightText: Copyright (c) 2020 Bryan Siepert for Adafruit Industries
# SPDX-License-Identifier: MIT
import time
import board

import adafruit_pcf8591.pcf8591 as PCF
from adafruit_pcf8591.analog_in import AnalogIn

################ AnalogIn Example #####################
#
# This example shows how to use the AnalogIn class provided
# by the library by creating an AnalogIn instance and using
# it to measure the voltage at the first ADC channel input
#
# Wiring:
# Connect a voltage source to the first ADC channel, in addition to the
# normal power and I2C connections. The voltage level should be between 0V/GND and VCC
#
########################################

i2c = board.I2C()
pcf = PCF.PCF8591(i2c)

pcf_in_0 = AnalogIn(pcf, PCF.A0)
while True:
    raw_value = pcf_in_0.value
    scaled_value = (raw_value / 65535) * pcf_in_0.reference_voltage

    print("Pin 0: %0.2fV" % (scaled_value))
    print("")
    time.sleep(1)



# SPDX-FileCopyrightText: Copyright (c) 2020 Bryan Siepert for Adafruit Industries
#
# SPDX-License-Identifier: MIT

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
"""
`analog_out`
==============================
AnalogOut for setting the included DAC to a given voltage.

* Author(s): Bryan Siepert
"""


class AnalogOut:
    """AnalogIn Mock Implementation for ADC Reads."""

    def __init__(self, pcf, dac_pin=0):
        """AnalogIn

        :param pcf: The pcf object.
        :param ~digitalio.DigitalInOut DAC pin: Required pin must be P4

        """
        self._pcf = pcf
        if dac_pin != 0:
            raise AttributeError("DAC pin must be adafruit_pcf8591.pcf8591.DAC_PIN")
        self._pin_setting = dac_pin
        self._pcf.dac_enabled = True

    @property
    def value(self):
        """Returns the currently set value of the DAC pin as an integer."""
        return self._value

    @value.setter
    def value(self, new_value):  # this may have to scale from 16-bit
        if new_value < 0 or new_value > 65535:
            raise ValueError("value must be a 16-bit integer from 0-65535")

        if not self._pcf.dac_enabled:
            raise RuntimeError(
                "Underlying DAC is disabled, likely due to callint `deinit`"
            )
        # underlying sensor is 8-bit, so scale accordingly
        self._pcf.write(new_value >> 8)
        self._value = new_value

    def deinit(self):
        """Disable the underlying DAC and release the reference to the PCF8591.
        Create a new AnalogOut to use it again."""
        self._pcf.dac_enabled = False
        self._pcf = None
